--------------------------------------------------------------------------------
--
-- Módulo SGA de Testes para Grid
--
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Chaves padrões e chaves disponíveis da biblioteca
--------------------------------------------------------------------------------

NODES = {}
-- NODES["Retina-MacBook-Pro"] = {
NODES["lfpina"] = {
  csbase_num_processors = 4,
  csbase_memory_ram_info_mb = 16384,
  csbase_memory_swap_info_mb = 524288,
  csbase_job_control_actions = {},
  csbase_load_perc = 0.50,
  csbase_memory_ram_free = 8192,
  csbase_memory_swap_free = 262144,
  csbase_number_of_jobs = 0,
}
NODES["testNode2"] = {
  csbase_num_processors = 4,
  csbase_memory_ram_info_mb = 32768,
  csbase_memory_swap_info_mb = 131072,
  csbase_job_control_actions = {},
  csbase_load_perc = 0.90,
  csbase_memory_ram_free = 3276,
  csbase_memory_swap_free = 13107,
  csbase_number_of_jobs = 0,
}
NODES["testNode3"] = {
  csbase_num_processors = 4,
  csbase_memory_ram_info_mb = 16384,
  csbase_memory_swap_info_mb = 2097152,
  csbase_job_control_actions = {},
  csbase_load_perc = 0.20,
  csbase_memory_ram_free = 13107,
  csbase_memory_swap_free = 1677721,
  csbase_number_of_jobs = 0,
}
NODES["testNode4"] = {
  csbase_num_processors = 2,
  csbase_memory_ram_info_mb = 65536,
  csbase_memory_swap_info_mb = 131072,
  csbase_job_control_actions = {},
  csbase_load_perc = 0.10,
  csbase_memory_ram_free = 58982,
  csbase_memory_swap_free = 117964,
  csbase_number_of_jobs = 0,
}

-- Chaves padrões de configuração da biblioteca
_libDefaltConfigKeys = {
  "csbase_num_processors",
  "csbase_memory_ram_info_mb",
  "csbase_memory_swap_info_mb",
  "csbase_job_control_actions",
}

-- Chaves padrões de configuração definidas pelo usuário
_userDefaltConfigKeys = _libDefaltConfigKeys

-- Chaves de configuração disponíveis
-- O valor true indica que é um chave obrigatório e false indica que é opcional
_allConfigKeys = {
  csbase_num_processors = true,
  csbase_memory_ram_info_mb = true,
  csbase_memory_swap_info_mb = true,
  csbase_job_control_actions = true,
}

-- Chaves padrões de informações dinâmicas da biblioteca
_libDefaltInfoKeys = {
  "csbase_load_perc",
  "csbase_memory_ram_free",
  "csbase_memory_swap_free",
  "csbase_number_of_jobs",
}

-- Chaves padrões de informações dinâmicas definidas pelo usuário
_userDefaltInfoKeys = _libDefaltInfoKeys

-- Chaves de informações dinâmicas disponíveis
-- O valor true indica que é um chave obrigatório e false indica que é opcional
_allInfoKeys = {
  csbase_load_perc = true,
  csbase_memory_ram_free = true,
  csbase_memory_swap_free = true,
  csbase_number_of_jobs = true,
}

-- Chaves padrões de informações de processo da biblioteca
_libDefaltProcessKeys = {
  "csbase_command_pid",
  "csbase_command_string",
  "csbase_command_exec_host",
  "csbase_command_state",
  "csbase_command_memory_ram_size_mb",
  "csbase_command_memory_swap_size_mb",
  "csbase_command_cpu_perc",
  "csbase_command_time_sec",
  "csbase_command_wall_time_sec",
  "csbase_command_user_time_sec",
  "csbase_command_system_time_sec",
}

-- Chaves de informações de processos definidas pelo usuário
_userDefaltProcessKeys = _libDefaltProcessKeys

-- Chaves de informações de processos disponíveis
-- O valor true indica que é um chave obrigatório e false indica que é opcional
_allProcessKeys = {
  csbase_command_pid = true,
  csbase_command_string = true,
  csbase_command_exec_host = true,
  csbase_command_state = true,
  csbase_command_memory_ram_size_mb = true,
  csbase_command_memory_swap_size_mb = true,
  csbase_command_cpu_perc = true,
  csbase_command_time_sec = true,
  csbase_command_wall_time_sec = true,
  csbase_command_user_time_sec = true,
  csbase_command_system_time_sec = true,
}

open = function()
  return true
end -- function open

close = function()
  servermanager._nodes = nil
end -- function close

setDefaultKeys = function(keys)
  _userDefaltConfigKeys = {}
  _userDefaltInfoKeys = {}
  _userDefaltProcessKeys = {}
  for _,v in ipairs(keys) do
    -- Config keys
    if(_allConfigKeys[v]) then
      table.insert(_userDefaltConfigKeys, v)
    end -- Config keys
    -- Info keys
    if(_allInfoKeys[v]) then
      table.insert(_userDefaltInfoKeys, v)
    end -- Info keys
    -- Process keys
    if(_allProcessKeys[v]) then
      table.insert(_userDefaltProcessKeys, v)
    end -- Process keys
  end
end -- function setDefaultKeys

getAllKeys = function()
  local allKeys = {}
  for k,_ in pairs(_allConfigKeys) do
    table.insert(allKeys, k)
  end

  for k,_ in pairs(_allInfoKeys) do
    table.insert(allKeys, k)
  end

  for k,_ in pairs(_allProcessKeys) do
    table.insert(allKeys, k)
  end
end -- function getAllKeys

getConfiguration = function(server, keys)
  local configKeys
  if not keys or #keys == 0 then
    configKeys = _userDefaltConfigKeys
  else
    configKeys = keys
  end

  local node = servermanager._getNode(server)
  local configMap = {}
  for _, k in ipairs(configKeys) do
    configMap[k] = node[k]
  end

  return configMap, nil
end -- function getConfiguration

getInfo = function(server, keys, maxold)
  local infoKeys
  if not keys or #keys == 0 then
    infoKeys = _userDefaltInfoKeys
  else
    infoKeys = keys
  end

  _update(server, maxold)

  local node = servermanager._getNode(server)
  local infoMap = {}
  for _, k in ipairs(infoKeys) do
    infoMap[k] = node[k]
  end

  return infoMap, nil
end -- function getInfo

executeCommand = function(id, command, extraParams, server)
  local handle = {}
  handle.command = command
  handle.host = _pickNode()
  handle.time_file = os.tmpname()
  handle.cmdid = id
  handle.last_update = 0
  handle.pid = servermanager._createProcess(handle.host)

  servermanager.writeCmd( "Tentando execução:\n\t"..id.."\n" )

  return handle, nil
end -- function executeCommand

control = function(handle, action, childId)
-- @TODO Pegar o enumerador da IDL para validar
  if action ~= "TERMINATE" then
    return nil, "Ação não suportada"
  end

  if type(handle) ~= "table" or
    not handle.pid or not tonumber(handle.pid) then
    return nil, "Handle inválido."
  end

  local process = servermanager._processes[handle.pid]
  process.state = servermanager.FINISHED

  return 1
end -- function control

getCommandPersistentData = function(handle)
  if type(handle) ~= "table" or not handle.pid
     or not tonumber(handle.pid) then
     return nil, "Handle inválido."
  end

  if (not handle.pscommand) then
     -- handle.pscommand = @TODO definir uma string
  end
  return { pid = tonumber(handle.pid),
           cmdid = handle.cmdid,
           time_file = handle.time_file,
           pscommand = handle.pscommand,
           host = handle.host,
         }, nil
end -- function getCommandPersistentData

retrieveCommandHandle = function(pdata)
  if type(pdata) ~= "table" or not pdata.pid or
    not tonumber(pdata.pid) then
    return nil, "Dados inválidos."
  end

  if not servermanager._processes[pdata.pid] then
    return nil, "Comando não encontrado."
  end

  pdata.last_update = 0

  return pdata, nil
end -- function retrieveCommandHandle

getCommandInfo = function(handle, keys, maxold)
  if type(handle) ~= "table" or not handle.cmdid then
    return nil, "Handle inválido."
  end

  _update(handle.host, maxold)

  local time, err = _getCommandTimeUsage(handle, maxold)
  if err then
    time = {}
    time.elapsed = 0
    time.user = 0
    time.system = 0
  end

  local mem, err = _getCommandMemoryLoad(handle, maxold)
  if err then
    mem = {}
    mem.ram = 0
    mem.swap = 0
  end

  if not keys or #keys == 0 then
    keys = _userDefaltProcessKeys
  end -- if no keys

  local cmdInfoMap = {}

  for _, k in ipairs(keys) do
    if k == "csbase_command_id" then
      cmdInfoMap[k] = _getCommandId(handle)
    elseif k == "csbase_command_pid" then
      cmdInfoMap[k] = _getCommandPid(handle, maxold)
    elseif k == "csbase_command_ppid" then
      cmdInfoMap[k] = 0
    elseif k == "csbase_command_string" then
      cmdInfoMap[k] = handle.command
    elseif k == "csbase_command_exec_host" then
      cmdInfoMap[k] = _getCommandExecHost(handle, maxold)
    elseif k == "csbase_command_state" then
      cmdInfoMap[k] = _getCommandStatus(handle, maxold)
    elseif k == "csbase_command_memory_ram_size_mb" then
      cmdInfoMap[k] = mem.ram
    elseif k == "csbase_command_memory_swap_size_mb" then
      cmdInfoMap[k] = mem.swap
    elseif k == "csbase_command_cpu_perc" then
      cmdInfoMap[k] = _getCommandCPULoad(handle, maxold)
    elseif k == "csbase_command_cpu_time_sec" then
      cmdInfoMap[k] = 0
    elseif k == "csbase_command_wall_time_sec" then
      cmdInfoMap[k] = time.elapsed
    elseif k == "csbase_command_user_time_sec" then
      cmdInfoMap[k] = time.user
    elseif k == "csbase_command_system_time_sec" then
      cmdInfoMap[k] = time.system
    end
  end

  return cmdInfoMap, nil
end -- function getCommandInfo

releaseCommandResources = function(handle)
   return 1
end -- function releaseCommandResources

getJobsInfo = function()
  return nil
end -- function getJobsInfo

getNodes = function()
  local nodes = {}
  for k, _ in pairs(servermanager._nodes) do
    table.insert(nodes, k)
  end
  return nodes, nil
end -- function getNodes

setHistoric = function(historic)
  enableHistoric = historic
end -- function setHistoric

getHistoric = function()
  return enableHistoric
end -- function getHistoric

------------------------------------------------------------------------------
--
-- Especificação das funções privadas
--
------------------------------------------------------------------------------
writeMsg = function(str)
  local msg = "SGAD ["..os.date("%d/%m/%y - %H:%M:%S").."] - "..str
  print(msg)
  io.flush()
end -- function writeMsg

-- writeError = writeMsg
-- writeCmd = writeMsg

_buildErrMsg = function(keys)
  if #keys < 1 then
    return nil
  else
    local errorMessage =
      "Error while retriving configuration for the following keys : ["
    for i, k in ipairs(keys) do
      if i < #keys then
        errorMessage = errorMessage .. k .. ", "
      else
        errorMessage = errorMessage .. k .. "]."
      end
    end
    return errorMessage
  end
end -- function _buildErrMsg

_getNode = function(server)
  return servermanager._nodes[server]
end -- function _getNode

_update = function(node, freq)
  local nodeInfo = servermanager._getNode(node)
  -- if freq and (servermanager.now() - nodeInfo.last_update) < freq then
  --   return nodeInfo
  -- end

  -- nodeInfo.last_update = servermanager.now()

  for pid, process in pairs(servermanager._processes) do
    local deltaupdate = servermanager.now() - process.lastupdate
    local timeusage = process.timeusage
    timeusage.system = timeusage.system + deltaupdate * process.cpuload
    timeusage.user = timeusage.user + deltaupdate
    timeusage.elapsed = timeusage.elapsed + deltaupdate
    process.lastupdate = servermanager.now()

    local p = math.random()
    local node = servermanager._getNode(process.host)
    if process.state == servermanager.RUNNING then
      if p > 0.9 then
        node.csbase_load_perc = node.csbase_load_perc - process.cpuload
        if node.csbase_load_perc < 0 then
          node.csbase_load_perc = 0
        end
        node.csbase_memory_ram_free = node.csbase_memory_ram_free + process.memoryload.ram
        if node.csbase_memory_ram_free > node.csbase_memory_ram_info_mb then
          node.csbase_memory_ram_free = node.csbase_memory_ram_info_mb
        end
        node.csbase_memory_swap_free = node.csbase_memory_swap_free + process.memoryload.swap
        if node.csbase_memory_swap_free > node.csbase_memory_swap_info_mb then
          node.csbase_memory_swap_free = node.csbase_memory_swap_info_mb
        end
        node.csbase_number_of_jobs = node.csbase_number_of_jobs - 1
        if node.csbase_number_of_jobs < 0 then
          node.csbase_number_of_jobs = 0
        end

        process.state = servermanager.FINISHED
        process.cpuload = 0
        process.memoryload.ram = 0
        process.memoryload.swap = 0
      end
    -- elseif process.state == servermanager.NOT_RESPONDING then
    elseif process.state == servermanager.WAITING then
      if p > 0.2 then
        process.state = servermanager.RUNNING
        process.cpuload = math.random() * node.csbase_load_perc
        process.memoryload.ram = math.random() * node.csbase_memory_ram_free
        process.memoryload.swap = math.random() * node.csbase_memory_swap_free

        node.csbase_load_perc = node.csbase_load_perc + process.cpuload
        if node.csbase_load_perc > 1 then
          node.csbase_load_perc = 1
        end
        node.csbase_memory_ram_free = node.csbase_memory_ram_free - process.memoryload.ram
        if node.csbase_memory_ram_free < 0 then
          node.csbase_memory_ram_free = 0
        end
        node.csbase_memory_swap_free = node.csbase_memory_swap_free - process.memoryload.swap
        if node.csbase_memory_swap_free < 0 then
          node.csbase_memory_swap_free = 0
        end
        -- node.csbase_number_of_jobs = node.csbase_number_of_jobs + 1
      end
    else
      -- Se é FINISHED nada para fazer
    end
  end
end -- function _update

_createProcess = function(nodename)
  local process = {
    pid = math.random(1000, 9999),
    state = servermanager.WAITING,
    cpuload = 0,
    memoryload = {
      ram = 0,
      swap = 0
    },
    timeusage = {
      user = 0,
      system = 0,
      elapsed = 0,
    },
    lastupdate = servermanager.now(),
    host = nodename,
  }

  if not servermanager._processes[process.pid] then
    servermanager._processes[process.pid] = process
    local node = servermanager._getNode(nodename)
    node.csbase_number_of_jobs = node.csbase_number_of_jobs + 1
  else
    return _createProcess(node)
  end

  return process.pid
end -- function _createProcess

_pickNode = function()
  local nodes = servermanager.getNodes()
  return nodes[math.random(#nodes)]
end

-- @TODO Rever a lista
_getCommandActions = function()
  return {
   "SUSPEND",
   "RESUME",
   "HOLD",
   "RELEASE",
   "TERMINATE"
  }
end

_getCommandId = function(handle)
  if type(handle) ~= "table" or not handle.cmdid then
    return nil, "Handle inválido."
  end

  return handle.cmdid, nil
end -- function _getCommandId

_getCommandPid = function(handle)
  if type(handle) ~= "table"
    or not handle.pid
    or not tonumber(handle.pid) then

    return nil, "Handle inválido."
  end

  return handle.pid, nil
end -- function _getCommandPid

_getCommandExecHost = function(handle, maxold)
  if type(handle) ~= "table" then
    return nil, "Handle inválido."
  end

  if handle.host then
    return handle.host, nil
  end

  return nil, "host unknown"
end -- function _getCommandExecHost

_getCommandStatus = function(handle, maxold)
  if type(handle) ~= "table" or
    not handle.pid or not tonumber(handle.pid) or
    not handle.last_update or not tonumber(handle.last_update) then
    return nil, "Handle inválido."
  end

  local process = servermanager._processes[handle.pid]

  return process.state, nil
end -- function _getCommandStatus

_getCommandCPULoad = function(handle, maxold)
  if type(handle) ~= "table" or
    not handle.pid or not tonumber(handle.pid) or
    not handle.last_update or not tonumber(handle.last_update) then
    return nil, "Handle inválido."
  end

  local process = servermanager._processes[handle.pid]

  return process.cpuload, nil
end -- function getcommandcpuload

_getCommandMemoryLoad = function(handle, maxold)
  if type(handle) ~= "table" or
    not handle.pid or not tonumber(handle.pid) or
    not handle.last_update or not tonumber(handle.last_update) then
    return nil, "Handle inválido."
  end

  local process = servermanager._processes[handle.pid]

  return process.memoryload, nil
end -- function _getCommandMemoryLoad

_getCommandTimeUsage = function(handle, maxold)
  local errcode = servermanager.ERROR_CODE
  if type(handle) ~= "table" or
    not handle.pid or not tonumber(handle.pid) or
    not handle.last_update or not tonumber(handle.last_update) then
    return nil, "Handle inválido."
  end

  local process = servermanager._processes[handle.pid]

  return process.timeusage, nil
end -- function _getCommandTimeUsage

--------------------------------------------------------------------------------
servermanager = {
  -- Private
  NO_SERVER                 = "Servidor não existente",
  _nodes                    = NODES,
  _processes                = {},
  _getCommandFinalTime      = _getCommandFinalTime,
  _getNode                  = _getNode,
  _update                   = _update,
  _createProcess           = _createProcess,
  -- Public
  now                       = now,
  sleep                     = sleep,
  writeCmd                  = writeMsg,
  writeError                = writeMsg,
  writeMsg                  = writeMsg,
  -- Funções de controle do módulo:
  open                      = open,
  close                     = close,
  -- Funções de consulta à configuração de servidor:
  setDefaultConfigKeys      = setDefaultConfigKeys,
  getAllConfigKeys          = getAllConfigKeys,
  getConfiguration          = getConfiguration,
  -- Funções de monitoração de servidor:
  setDefaultInfoKeys        = setDefaultInfoKeys,
  getAllInfoKeys            = getAllInfoKeys,
  getInfo                   = getInfo,
  -- Funções de execução, monitoração e controle de processos:
  control                   = control,
  executeCommand            = executeCommand,
  getCommandInfo            = getCommandInfo,
  getCommandPersistentData  = getCommandPersistentData,
  releaseCommandResources   = releaseCommandResources,
  retrieveCommandHandle     = retrieveCommandHandle,
  getJobsInfo               = getJobsInfo,
  getNodes                  = getNodes,
  -- Funções de log de histórico
  setHistoric               = setHistoric,
  getHistoric               = getHistoric,
  -- Constantes do módulo:
  RUNNING                   = 0,
  NOT_RESPONDING            = 1,
  WAITING                   = 2,
  FINISHED                  = 3,
  LITTLE_ENDIAN             = 0,
  BIG_ENDIAN                = 1,
  ERROR_CODE                = -1,
}